require 'fileutils'
require 'find'
require 'net/http'
require 'timeout'
require 'uri'
require 'matrix'

#inherits from math.rb's Vector
class Vector
	def dotProduct(p)
		x=0
		y=1
		return self[x] * p[x] + self[y] * p[y]
	end
	def angle(v2)
		v1 = self
		dot = v1.dotProduct(v2)
		length_product = (v1.r*v2.r)
		if ((dot-length_product).abs < 0.000001)
			return 0.0
		else
			#$stderr.print(sprintf("length_product: %f dot: %f\n", length_product, dot))
			#$stderr.print(sprintf("v1: %s v2: %s v1.r: %f v2.r: %f\n", v1.to_s, v2.to_s, v1.r, v2.r))
		end
		#if (dot/length_product >= 1)
		#end
		angle_radians = Math.acos(dot/length_product)
		return (180/Math::PI)*angle_radians
	end
end


module OSMModel
def debug(string)
	#$stderr.puts string
end
module_function :debug

def hash_to_tags(h)
	s = []
	if ! h.empty?
		s.push("\n");
	end
	h.each_pair do |key, value|
		if (value.nil? || (value.to_s.length == 0))
			next
		end
		value.gsub(/\'/, "\\'")
		value.gsub(/\&/, "amp")
		s += [ "\t<tag k=\"#{key}\" v=\"#{value}\"/>\n" ]
	end
	s
end
module_function :hash_to_tags
$way_types = ["highway", "railway", "power", "man_made", "route"]
def tags_to_feature_type(tags)
	$way_types.each do |way_type|
		next if tags[way_type.to_s].nil?
		return way_type
	end
	return nil
end
module_function :tags_to_feature_type

$way_cache
$ways_with_endpoint_at_node = Hash.new
$combine_success = 0
$combine_fail = 0
def coalesce_way(new_way)
	success = false
	possible_ways = []
	if !$ways_with_endpoint_at_node[new_way.first_node].nil?
		possible_ways.concat($ways_with_endpoint_at_node[new_way.first_node])
	end
	if !$ways_with_endpoint_at_node[new_way.last_node].nil?
		possible_ways.concat($ways_with_endpoint_at_node[new_way.last_node])
	end
	possible_ways.each do |way|
		if ! way.name.eql?(new_way.name)
			next;
		end
    		if way.combine_with(new_way)
			# the new way may have connected two
			# existing ways, we need to recursively
			# continue looking for other completed
			# ways
			$combine_success += 1
			new_way.remove_from_endpoint_cache;
			way.refresh_endpoint_cache_entry()
			coalesce_way(way)
			success = true
			break;
		else
			$combine_fail += 1
		end
	end
	return success
end
module_function :coalesce_way
def create_way(coalesce, segs, tags = {})
	if $way_cache.nil?
		$way_cache = Hash.new
	end
	key = tags["name"]
 	if $way_cache[key].nil?
    		$way_cache[key] = []
    	end
    	new_way = OSMModel::Way.new(segs, tags)
	$way_cache[key].push(new_way)
  	if ($way_cache[key].size > 40)
		#$stderr.printf("there are %d ways with key: '%s'\n",  $way_cache[key].size, key);
	end
	if coalesce
		success = OSMModel::coalesce_way(new_way)
		if ! success
			new_way.refresh_endpoint_cache_entry()
		end
	end
	return
	$way_cache[key].each do |way|
		break if ! coalesce
    		if way.combine_with(new_way)
			$combine_success += 1
			# "new_way" was combined into another
			# now we need to get rid of it
    			$way_cache[key].delete(new_way)
			new_way.remove_from_endpoint_cache;
	    		new_way = way
			way.refresh_endpoint_cache_entry()
		else
			$combine_fail += 1
		end
    	end
end
module_function :create_way
def get_all_ways
	return $way_cache
end
module_function :get_all_ways
def print_stats
	$stderr.printf("   combine_fail: %d\n", $combine_fail);
	$stderr.printf("combine_success: %d\n", $combine_success);
end
module_function :print_stats

$node_cache
def new_node(point, feature_type, tags)
        node = nil
	lat = point[1]
	lon = point[0]
        loop = 0;
	if $node_cache.nil?
		$node_cache = Hash.new
	end
	#stderr.print(sprintf("new_node() feature_type: '%s'\n", feature_type))
        key = lat+lon
        while $node_cache.has_key?(key)
                node = $node_cache[key]
       		if node.feature_type == feature_type &&
		   node.distance(lat,lon) < 0.0000001
			#print "found close node\n"
			if node.combine_tags(tags)
	                	#stderr.puts "cache hit"
				return node
			end
		else
			#printf "actual hash collision: %f,%f vs %f,%f\n", lat, lon, node.lat, node.lon
		end
                key += lat+lon
                loop+=1
                #stderr.puts "loop nr: #{loop}" if loop > 2
        end
        node = Node.new(point, feature_type, tags)
        while $node_cache.has_key?(key)
                key += lat+lon
        end
        $node_cache[key] = node
        node
end
module_function :new_node
$node_ids = -1
class Node
	attr	:lat
	attr	:lon
	attr_accessor :osmid
	attr_accessor :feature_type
	attr_accessor :first_node
	attr_accessor :last_node
        def initialize point, feature_type, tags={}
	      @visible="true"
	      @lat = point[1]
	      @lon = point[0]
	      @feature_type = feature_type
	      @tags = tags
	      @osmid = ($node_ids-=1)
	end
	def -(node)
		xdiff = self.lon.to_f - node.lon.to_f
		ydiff = self.lat.to_f - node.lat.to_f
		xdiff *= 100.0
		ydiff *= 100.0
		#printf "lat/lon: %f/%f\n", self.lat, self.lon
		#printf "xdiff: %f\n", xdiff
		#printf "ydiff: %f\n", ydiff
		return Vector[xdiff, ydiff]
	end
	def combine_tags(tags)
		replacements = Hash.new
		tags.each_pair do |key, value|
			next if @tags[key] == value
			if key == "tiger:tzid" ||
			   key == "tiger:tlid"
				replacements[key] = @tags[key] + ":" + value
				next
			end
			OSMModel::debug(sprintf("incompatible tag: '%s' values: '%s' '%s'\n", key, value, @tags[key]))
			return false
		end
		replacements.each_pair do |key, value|
			@tags[key] = replacements[key]
		end
		replacements.clear
		return true
	end
	def same_position_as(node)
		if node.lat == self.lat &&
		   node.lon == self.lon
			return 1
		end
		return 0
	end
	def distance(lat, lon)
		if lat == self.lat &&
		   lon == self.lon
			return 0
		end
		latdiff = (lat - self.lat).abs
		londiff = (lon - self.lon).abs
		return Math.sqrt(latdiff*latdiff + londiff*londiff)
	end
	def to_xml
		return [] if @printed
		@printed = true
		return [ "<node id=\'#{@osmid}\' visible=\'true\' lat=\'#{@lat}\' lon=\'#{@lon}\'>" ] +
			OSMModel.hash_to_tags(@tags) + [ "</node>\n" ]
	end
	def id
		@osmid
	end
	def to_s
		"[#{@lat}, #{@lon}]"
	end
end
 
# Create a new object
# g = Greeter.new("world")
#  
#  # Output "Hello World!"
#  g.salute
$segment_ids = -1000000;
class Segment
	attr :to
	attr :from
	attr :tags
	attr_accessor :osmid
	def initialize(from, to, tags = {})
		@from = from
		@to = to
		@tags = tags
		@printed = false
		#OSMModel::debug("created segment: " + self.to_s)
		@osmid = ($segment_ids-=1)
  	end
	def vector
		return (from - to)
	end
	def length
		return vector.r
	end
	def comes_before(seg)
		if self.to == seg.from
			return true
		end
		return false
	end
	def reverse
		tmp = @from
		@from = @to
		@to = tmp
	end
	def to_xml
		return [] if @printed
		@printed = true
		ret = from.to_xml() + to.to_xml();
		ret += [ "<segment from=\'#{from.id}\' to=\'#{to.id}\' id=\'#{self.id}\' visible=\'true\'>" ]
		ret += OSMModel.hash_to_tags(@tags)
		ret += [ "</segment>\n" ]
	end
	def id
		@osmid
	end
	def to_s
		"segment:\{id:#{@osmid} from:#{@from.to_s} to:#{@to.to_s}\}"
	end
end
 
def create_segment(from_node, to_node, tags = {})
	return Segment.new(from_node, to_node, tags)
end

$way_ids = -2000000;
class Way
	attr :segs
	attr :tags
	attr_accessor :osmid
	attr_accessor :first_node
	attr_accessor :last_node
	attr_accessor :combined
	def initialize(segs, tags = {})
		@osmid = ($way_ids-=1)
		@segs = []
		#$stderr.printf("adding segments to new way: %d\n", @osmid)
		@tags = tags
		segs.each do |seg|
			#$stderr.printf("adding segment...\n");
			self.add_segment(seg)
		end
		#$stderr.printf("done adding segments to new way: %d\n", @osmid)
  	end
	def combine_tags(tags)
		replacements = Hash.new
		tags.each_pair do |key, value|
			next if @tags[key] == value
			if key == "tiger:tlid"
				replacements[key] = @tags[key] + ":" + value
				next
			end
			OSMModel::debug(sprintf("incompatible tag: '%s' values: '%s' '%s'\n", key, value, @tags[key]))
			return false
		end
		replacements.each_pair do |key, value|
			@tags[key] = replacements[key]
		end
		replacements.clear
		return true
	end
	def reverse
		@segs.each do |seg|
			seg.reverse
		end
		@segs.reverse!
		tmp = @__first_node
		@__first_node = @__last_node
		@__last_node = tmp
	end
	def first_node
		return @segs.first.from
	end
	def last_node
		return @segs.last.to
	end
	def name
		return @tags["name"]
	end
	def remove_from_endpoint_cache_node(node, msg)
		#$stderr.printf("%d removing from endpoint cache %s %s\n", @osmid, msg, node.to_s)
		if node.nil?
			return
		end
		#$stderr.printf("%d remove_from_endpoint_cache_node(%s, %s)\n", @osmid, node.to_s, msg)
		array = $ways_with_endpoint_at_node[node]
		# If the first and last nodes are the same, just doing a array.delete(self)
		# will delete both of them.  Not good.
		array.delete_at(array.index(self))
		#if !array.index(self).nil?
		#	$stderr.printf("way: %d still in array under node: %s\n", @osmid, node.to_s)
		#end
		if array.empty?
		#	$stderr.printf("%d deleting array for %s: %s\n", @osmid, msg, node.to_s)
			$ways_with_endpoint_at_node.delete(node)
		#else
		#	$stderr.printf("%d array for %s has %d items\n", @osmid, msg, array.length)
		end
	end
	def remove_from_endpoint_cache
		remove_from_endpoint_cache_node(@__first_node, "(first node)")
		remove_from_endpoint_cache_node(@__last_node, "(last node)")
		#$ways_with_endpoint_at_node.each_pair do |node, way_array|
		#	way_array.each do |way|
		#		if way == self
		#			$stderr.printf("%d: uh oh, way still in cache under node: %s\n", @osmid, node)
		#			exit 1;
		#		end
		#	end
		#end
		@__first_node = nil
		@__last_node = nil
	end
	def replace_in_endpoint_cache(node, newnode)
		if !node.nil?
			array = $ways_with_endpoint_at_node[node]
			array.delete_at(array.index(self))
			if $ways_with_endpoint_at_node[node].length == 0
				$ways_with_endpoint_at_node.delete(node)
			end
		end
		if $ways_with_endpoint_at_node[newnode].nil?
			$ways_with_endpoint_at_node[newnode] = []
		end
		$ways_with_endpoint_at_node[newnode].push(self);
		#$stderr.printf("\n%d replace_in_endpoint_cache(%s, %s)\n", @osmid, node.to_s, newnode.to_s)
	end
	def refresh_endpoint_cache_entry
		if @__first_node != @segs.first.from
			self.replace_in_endpoint_cache(@__first_node, @segs.first.from)
			@__first_node = @segs.first.from;
		end
		if @__last_node != @segs.last.to
			self.replace_in_endpoint_cache(@__last_node, @segs.last.to)
			@__last_node = @segs.last.to;
		end
		#if @__last_node == @__first_node
		#	$stderr.printf("%d: uh oh, circular way\n", @osmid)
		#end
	end
	def __add_segment(newseg)
		result = false
		if @segs.empty?
			OSMModel::debug("added " + newseg.to_s  + " to empty way #{@osmid}")
			@segs.push(newseg)
			result = true
		elsif newseg.comes_before(@segs.first)
			OSMModel::debug("added "+newseg.to_s+" to head of way #{@osmid} before " + @segs.first.to_s)
			@segs.unshift(newseg)
			result = true
		elsif @segs.last.comes_before(newseg)
			OSMModel::debug("added "+newseg.to_s+" to tail of way #{@osmid} after " + @segs.last.to_s)
			@segs.push(newseg)
			result = true
		end
		return result
	end
	def add_segment(newseg)
		if __add_segment(newseg)
			return true
		end
		# try the segment in the opposite direction
		newseg = Segment.new(newseg.to, newseg.from, newseg.tags)
		return __add_segment(newseg)
	end
	def combine_with(way)
		return if way == self
		debug = false
		#if (self.name.eql?("I-75")) 
		#	$stderr.printf("\ncombine_with() debug\n")
		#	$stderr.printf("%d: self tlid: '%s'\n", self.osmid, self.tags["tiger:tlid"])
		#	$stderr.printf("%d:  way tlid: '%s'\n", way.osmid,  way.tags["tiger:tlid"])
		#	#debug = true
		#end
		# check tag compatibility here
		#way_segs = way.segs
		#way_segs_first = way_segs.first
		#way_segs_last = way_segs.last
		#segs_first = @segs.first
		#segs_last = @segs.last
		#$stderr.printf("%d way.segs.len: %d\n", way.osmid, way.segs.length)
		#$stderr.printf("%d self.segs.len: %d\n", self.osmid, self.segs.length)
		if ((way.segs.first.from == @segs.first.from) ||
		    (way.segs.last.to ==  @segs.last.to))
			#$stderr.printf("reversed\n") if debug
			way.reverse
		end
		combined = false
		max_angle = 60
		if @segs.last.to == way.segs.first.from
			#$stderr.printf("segs_last.to == way_segs_first.from\n") if debug
			v1 = @segs.last.vector
			v2 = way.segs.first.vector
			angle = v1.angle(v2)
			if (angle < max_angle)
				@segs.concat(way.segs)
				combined = true
			end
		elsif way.segs.last.to == @segs.first.from
			#$stderr.printf("way_segs_last.to == segs_first.from\n") if debug
			v1 = @segs.first.vector
			v2 = way.segs.last.vector
			angle = v1.angle(v2)
			if (angle < max_angle)
				@segs = way.segs | @segs
				combined = true
			end
		else
			#
		end
		if combined
			self.combine_tags(way.tags)
			way.segs.clear
			way.combined = true
			#$stderr.printf("combined %d into %d\n", way.osmid, self.osmid)
		end
		# if the passed in way is empty, we've succeeded
		#$stderr.printf("done\n") if debug
		return combined
	end
	def to_xml
		ret = []
		@segs.each do |seg|
			ret.concat(seg.to_xml)
		end
		ret += [ "<way id=\'#{@osmid}' visible=\'true\'>" ] +
		[ @segs.map { |seg| "\n\t<seg id=\'#{seg.id}\'/>" }.join ] +
		OSMModel.hash_to_tags(@tags) +
		[ "</way>\n" ]
	end
	def name
		return @tags["name"]
	end
	def id
		@osmid
	end
	def verify
		return
		last_seg = nil
		@segs.each do |seg|
			if !last_seg.nil?
				if last_seg.to != seg.from
					$stderr.puts "invalid way: " + self.to_s
					exit(1)
				end
			end
			last_seg = seg
		end
	end
	def to_s
		ret = "way:\{id:#{@osmid} segments: ["
		@segs.each do |seg|
			ret += "\n   " + seg.to_s + ","
		end
		ret += "]\}"
	end
end

end
