#extends to the TigerLine::Dataset class to have to_osm and to_osm_xml member functions.

require 'tiger'
require 'attr_map'
require 'cgi' #to escape srings
require 'dave_model'

module TigerLine

class Dataset
    def to_osm(coalesce)
      ret = []
      tzid_to_osm = {}

      GC.enable
      features_orig_size = @features.size
      printf "converting %d TIGER features into OSM objects...\n", @features.size
      nr = 0
      @features.each_pair do |tlid, feature|
	nr += 1
	if (nr%100 == 0)
      	  $stderr.write(sprintf("\rconverted %d/%d TIGER features to OSM (%5.2f%%)", nr, features_orig_size, (nr.to_f*100/features_orig_size)))
	end
        unless raw_tags = TIGER_TO_OSM[feature.cfcc.downcase.intern] then
	  @features.delete(tlid)
          next
        end
	way_tags = Hash.new
        osmpoints = []
	feature_type = OSMModel::tags_to_feature_type(way_tags)
	node_tags = Hash.new
	program_name = "tiger_import_dch_v0.4_#{Time.now.strftime("%Y%m%d")}"
        node_tags["source"] = program_name
	node_tags["tiger:tzid"] = feature.tzids
	node_tags["tiger:tlid"] = feature.tlid
	node_tags["tiger:tzide"] = feature.tzide
        feature.points.each do |point|
          osmpoints << OSMModel::new_node(point, feature_type, node_tags)
        end

	way_segs = []
        osmpoints[0..-2].each_with_index do |point, i|
		segment = OSMModel::Segment.new( point, osmpoints[i+1] )
		way_segs << segment
        end

        #name tags
        feature.names.each_with_index do |name, i|
          strname = "#{name[:fedirp]} #{name[:fename]} #{name[:fetype]} #{name[:fedirs]}".strip
          ord_suffix = ("_#{i}" if i>0) or ""
          way_tags["name#{ord_suffix}"] = "#{CGI.escapeHTML(strname)}"
          way_tags["tiger:name_direction_prefix#{ord_suffix}"] = "#{CGI.escapeHTML(name[:fedirp])}" if not name[:fedirp].empty?
          way_tags["tiger:name_base#{ord_suffix}"] = "#{CGI.escapeHTML(name[:fename])}" if not name[:fename].empty?
          way_tags["tiger:name_type#{ord_suffix}"] = "#{CGI.escapeHTML(name[:fetype])}" if not name[:fetype].empty?
          way_tags["tiger:name_direction_suffix#{ord_suffix}"] = "#{CGI.escapeHTML(name[:fedirs])}" if not name[:fedirs].empty?
        end
        #address range tags
        feature.address_ranges.each_with_index do |range, i|
          ord_suffix = ("_#{i}" if i>0) or ""
          #way_tags["from_address_right#{ord_suffix}"] = "#{range[:fraddr]}" if not range[:fraddr].empty?
          #way_tags["to_address_right#{ord_suffix}"] = "#{range[:toaddr]}" if not range[:toaddr].empty?
          #way_tags["from_address_left#{ord_suffix}"] ="#{range[:fraddl]}" if not range[:fraddl].empty?
          #way_tags["to_address_left#{ord_suffix}"] = "#{range[:toaddl]}" if not range[:toaddl].empty?
          way_tags["tiger:zip_left#{ord_suffix}"] = "#{range[:zipl]}" if not range[:zipl].empty?
          way_tags["tiger:zip_right#{ord_suffix}"] = "#{range[:zipr]}" if not range[:zipr].empty?
        end
        #preserved tiger fields
        way_tags["tiger:tlid"] = "#{tlid}"
        way_tags["tiger:cfcc"] = "#{feature.cfcc}"
        #misc tags
        way_tags["tiger:source"] = program_name
        way_tags["tiger:reviewed"] = "no"
        #cfcc tags
	way = OSMModel::create_way(coalesce, way_segs, way_tags)
	#way_segs.clear
	#way_tags.clear
	#node_tags.clear
	#features.delete(tlid)
      end
      printf "...done\n"
      return ret
    end
    def write_string_array(fd, array)
	array.each do |member|
		fd.write member
	end
    end

    def write_osm_xml(coalesce, fd)
      self.to_osm(coalesce)
      osmid = 1
      fd.write "<?xml version='1.0' encoding='UTF-8'?>\n"
      fd.write "<osm version='0.4' generator='JOSM'>\n"

      @ways = OSMModel::get_all_ways
      printf "writing %d ways into xml\n", @ways.size
      nr = 0
      ways_size = @ways.size
      print_each = ways_size/100
      if (print_each == 0)
	print_each = 1
      end     
      @ways.each_key do |name|
	if (nr%print_each == 0)
      	  $stderr.write(sprintf("done writing %d ways (%d%%)\r", nr, (nr*100/ways_size)))
	end
	nr+=1
      	@ways[name].each do |way|
        	way.segs.each do |seg|
			if seg.from.osmid==0 then 
				seg.from.osmid = (osmid+=1)
				write_string_array(fd, seg.from.to_xml)
			end
			if seg.to.osmid==0 then 
				seg.to.osmid = (osmid+=1)
				write_string_array(fd, seg.to.to_xml)
			end
			seg.osmid = (osmid+=1)
			write_string_array(fd, seg.to_xml)
		end
		way.osmid = (osmid+=1)
		write_string_array(fd, way.to_xml)
		#@ways[name].delete(way)
	end
	#@ways.delete(name)
      end
      $stderr.write(sprintf("wrote %d ways (%d%%)         \n", nr, (nr*100/ways_size)))
      fd.write  "</osm>"
    end
end

end
